#include <glad/glad.h>

#include <GLFW/glfw3.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>

#include "fio.h"
#include "gfx.h"

static void glfw_error_cb(int err, const char *description) {
    fprintf(stderr, "[GLFW] %u: %s\n", err, description);
}

static void make_window_windowed_fullscreen(GLFWwindow *window,
                                            GLFWmonitor *monitor) {
    const GLFWvidmode *mode = glfwGetVideoMode(monitor);
    glfwSetWindowMonitor(window,
                         monitor,
                         0,
                         0,
                         mode->width,
                         mode->height,
                         mode->refreshRate);
}

static void make_window_windowed(GLFWwindow *window, int w, int h) {
    glfwSetWindowMonitor(window, NULL, 0, 0, w, h, GLFW_DONT_CARE);
}

typedef struct {
    int f1;
    int esc;
} key_states;

int main(int argc, char **argv) {
    glfwSetErrorCallback(glfw_error_cb);
    if (!glfwInit()) {
        return 1;
    }
    printf("[GFLW] Version: %s\n", glfwGetVersionString());

    /* Create a windowed full screen window */
    glfwWindowHint(GLFW_CLIENT_API, GLFW_OPENGL_API);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 5);
    GLFWwindow *window = glfwCreateWindow(640, 480, "Voyage", NULL, NULL);
    if (!window) {
        fprintf(stderr, "[GFX] Window creation failed.\n");
        return 1;
    }
    glfwMakeContextCurrent(window);
    glfwSwapInterval(1);

    if (!gladLoadGL()) {
        fprintf(stderr, "[GFX] OpenGL load failed.\n");
        return 1;
    }

    vy_fio_config fio_config = {0};
    if (!vyInitFIO(&fio_config)) {
        fprintf(stderr, "[FIO] Init failed.\n");
        return 1;
    }

    if (!vyInitRenderer()) {
        fprintf(stderr, "[GFX] Init failed.\n");
        return 1;
    }
    printf("GL_MAX_SHADER_STORAGE_BUFFER_BINDINGS: %u\n",
           GL_MAX_SHADER_STORAGE_BUFFER_BINDINGS);
    printf("GL_MAX_UNIFORM_BUFFER_BINDINGS: %u\n",
           GL_MAX_UNIFORM_BUFFER_BINDINGS);

    key_states prev_keys = {0};
    int window_state     = 0;

    while (!glfwWindowShouldClose(window)) {
        /* Gather input */
        key_states keys;
        keys.f1  = glfwGetKey(window, GLFW_KEY_F1);
        keys.esc = glfwGetKey(window, GLFW_KEY_ESCAPE);

        if (keys.f1 && !prev_keys.f1) {
            if (window_state)
                make_window_windowed(window, 640, 480);
            else
                make_window_windowed_fullscreen(window,
                                                glfwGetPrimaryMonitor());
            window_state = !window_state;
        }

        if (keys.esc)
            glfwSetWindowShouldClose(window, GLFW_TRUE);

        int fbw, fbh;
        glfwGetFramebufferSize(window, &fbw, &fbh);
        glViewport(0, 0, fbw, fbh);
        glClearColor(0, 0, 0, 0);
        glClear(GL_COLOR_BUFFER_BIT);

        glfwSwapBuffers(window);
        glfwPollEvents();
        prev_keys = keys;
    }

    vyShutdownRenderer();
    vyShutdownFIO();

    glfwDestroyWindow(window);
    glfwTerminate();
    return 0;
}
